#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <pwd.h>
#include <sys/types.h>
#include <strings.h>
#include <string.h>
#include <search.h>

#include <sys/types.h>	/* basic system data types */
#include <sys/socket.h>	/* basic socket definitions */
#include <netinet/in.h>	/* sockaddr_in{} and other Internet defns */
#include <arpa/inet.h>	/* inet(3) functions */
#include <errno.h>
#include <netdb.h>

#include "stp_client.h"

#define	MAX_SERVER_HASH	200
#define	GROUP_KEYWORD	"G"
#define	SERVER_KEYWORD	"S"
#define	TOKENLEN	64
#define	HASHKEYLEN	(TOKENLEN+2)

typedef struct _server {
	char key[HASHKEYLEN];
	char name[TOKENLEN];
	char hostname[HOSTNAMELEN];
	int  port;
} SERVER;

typedef struct _group {
	char key[HASHKEYLEN];
	char name[TOKENLEN];
	int random;
	int nservers;
	char **servers;
} GROUP;

char *find_stp_config (char *config);
int process_stp_config (char *config);
int open_stpserver_at (char *hostname, int port);

/************************************************************************/
/*  find_stp_config						*/
/*	Find the stp server config file.				*/
/*	Return pathname on success, NULL on error.			*/
/************************************************************************/
char *find_stp_config (char *config)
{
    static char pathname[2048];
    struct passwd *pwent = NULL;
    sprintf (pathname, "./%s", config);
    if (access(pathname, R_OK) == 0) return (pathname);
    if ((pwent = getpwuid(getuid())) != NULL) {
	sprintf (pathname,"%s/%s",pwent->pw_dir,config);
	if (access(pathname, R_OK) == 0) return (pathname);
    }
    return (NULL);
}

/************************************************************************/
/*  stp_config							*/
/*	Find and process the stp server config file.			*/
/*	Return 0 on success, non-zero on error.				*/
/************************************************************************/
int stp_config (char *config)
{
    char *configpath = NULL;
    configpath = find_stp_config(config);
    if (configpath == NULL) return(-1);
    printf ("Using stp config file %s\n", configpath);
    return (process_stp_config(configpath));
}

#define	CONFIGERR(line,status) \
  { fprintf (stderr, "Error processing config line: %s", line); return (status); }

/************************************************************************/
/*  process_stp_config						*/
/*	Process the specified stp server config file.			*/
/*	Return 0 on success, non-zero on error.				*/
/************************************************************************/
int process_stp_config (char *config)
{
    FILE *fp;
    int status = -1;
    int i;
    char oline[MAXLINE], line[MAXLINE];
    char *cmd;
    ENTRY entry;
    int found;
    char *p;
    char default_key[128];

    if (config == NULL || strlen(config) == 0) return (status);
    if ((fp = fopen (config, "r")) == NULL) {
	fprintf (stderr, "Error: opening stp server config file %s for reading\n",
		 config);
	return (status);
    }

    if (hcreate (MAX_SERVER_HASH) == 0) 
        err(FATAL,"STP: Unable to create hash table for parsing config file\n");

    while (fgets(oline,MAXLINE,fp) != NULL) {
	if (oline[0] == '#') continue;	/* skip comments. */
	if (oline[0] == '*') continue;	/* skip comments. */
	strcpy (line, oline);
	cmd = strtok (line," \t\n");
	if (cmd == NULL) continue;
	if (strcasecmp(cmd,"server") == 0) {
	    char *name, *host, *port;
	    int pnum;
	    SERVER *server;
	    if ((name = strtok (NULL, " \t\n")) == NULL) CONFIGERR(oline,-1)
	    if ((host = strtok (NULL, " \t\n")) == NULL) CONFIGERR(oline,-1)
	    if ((port = strtok (NULL, " \t\n")) == NULL) CONFIGERR(oline,-1)
	    if ((pnum = atoi(port)) <= 0) CONFIGERR(oline,-1)
	    if ((server = (SERVER *)malloc(sizeof(SERVER))) == NULL) CONFIGERR(oline,-1)
	    strncpy(server->name,name,TOKENLEN);
	    server->name[TOKENLEN-1] = '\0';
	    strncpy(server->hostname,host,HOSTNAMELEN);
	    server->name[HOSTNAMELEN-1] = '\0';
	    sprintf (server->key, "%s:%s", SERVER_KEYWORD, server->name);
	    server->port = pnum;
	    /* Put server entry into hash table. */
	    entry.key = server->key;
	    entry.data = (void *)server;
	    if (hsearch (entry, FIND) != NULL)
		err(FATAL, "STP: server %s already defined\n", server->name);
	    if (hsearch (entry, ENTER) == NULL)
		err(FATAL, "STP: error entering server %s into hash table\n", server->name);
	}
	else if (strcasecmp(cmd,"group") == 0 || strcasecmp(cmd,"rgroup") == 0) {
	    char *name, *server;
	    GROUP *group;
	    if ((name = strtok (NULL, " \t\n")) == NULL) CONFIGERR(oline,-1)
	    if ((server = strtok (NULL, " \t\n")) == NULL) CONFIGERR(oline,-1)
	    if ((group = (GROUP *)malloc(sizeof(GROUP))) == NULL) CONFIGERR(oline,-1)
	    strncpy(group->name,name,TOKENLEN);
	    group->name[TOKENLEN-1] = '\0';
	    sprintf (group->key, "%s:%s", GROUP_KEYWORD, group->name);
	    group->random = (strcasecmp(cmd,"rgroup") == 0);
	    group->nservers = 0;
	    group->servers = NULL;
	    while (server != NULL) {
		/* Add the found server to this group. */
		char *servername = strdup(server);
		if ((group->servers = (char **)realloc(group->servers,(group->nservers+1)*sizeof(char *))) == NULL) 
		    CONFIGERR(oline,-1)
		group->servers[group->nservers++] = servername;
		/* Get the next server on the line. */
		server = strtok (NULL, " \t\n");
	    }
	    /* Put group entry into hash table. */
	    entry.key = group->key;
	    entry.data = (void *)group;
	    if (hsearch (entry, FIND) != NULL)
		err(FATAL, "STP: group %s already defined\n", group->name);
	    if (hsearch (entry, ENTER) == NULL)
		err(FATAL, "STP: error entering group %s into hash table\n", group->name);
	}
	else if (strcasecmp(cmd,"default") == 0) {
	    char *name;
	    if ((name = strtok (NULL, " \t\n")) == NULL) CONFIGERR(oline,-1)
	    default_connection = strdup(name);
	}
	else {
	    CONFIGERR(oline,-1)
	}
    }
    fclose(fp);

    /* Now ensure that we have a default group or server. */
    if (default_connection == NULL)
    	err(FATAL,"STP: No default found in config file\n");

    /* Look for a group with this name.  If found, use it. */
    sprintf (default_key, "%s:%s", GROUP_KEYWORD, default_connection);
    entry.key = default_key;
    if (hsearch (entry, FIND) == NULL) {
	sprintf (default_key, "%s:%s", SERVER_KEYWORD, default_connection);
	entry.key = default_key;
	if (hsearch (entry, FIND) == NULL)
	    err(FATAL,"STP: Cannot find default group or server: %s\n", default_connection);
    }
    status = 0;
    return (status);
}

/************************************************************************/
/*  open_stpserver:							*/
/*	Open a connection to STP server group (or server).		*/
/*	Return 0 on success, non-zero on failure.			*/
/*	Exit on fatal error.						*/
/************************************************************************/
int open_stpserver (char *name)
{
    ENTRY group_entry, server_entry, *ep;
    char group_key[128], server_key[128];
    if (name == NULL) {
	if (default_connection == NULL) 
	    err(FATAL, "STP: no default server defined in %s config file, or config file not found.\n",STP_CONFIG);
	name = default_connection;
    }
    sprintf (group_key, "%s:%s", GROUP_KEYWORD,name);
    sprintf (server_key, "%s:%s", SERVER_KEYWORD,name);
    group_entry.key = group_key;
    server_entry.key = server_key;

    /* First look for a group by this name. */
    if ((ep = hsearch(group_entry, FIND)) != NULL) {
	GROUP *group = (GROUP *)ep->data;
	int offset = 0;
	int i;
	/* Try all servers in this group until we connect to one. */
	if (group->random) {
	    unsigned int seed;
	    seed = time(NULL);
	    srandom(seed);
	    offset = random();
	    offset %= group->nservers;
/*::	    printf ("nservers=%d offset=%d\n", group->nservers, offset); */
	}
	for (i=0; i<group->nservers; i++) {
	    SERVER *server;
	    int j;
	    j = (i+offset) % group->nservers;
	    sprintf (server_key, "%s:%s", SERVER_KEYWORD, group->servers[j]);
	    server_entry.key = server_key;
	    if ((ep = hsearch(server_entry, FIND)) != NULL) {
		server = (SERVER *)(ep->data);
		if (open_stpserver_at (server->hostname, server->port) == 0) return (0);
	    }
	    else {
		err(FATAL,"STP: Warning - server %s in group %s not defined\n",group->servers[i],group->name);
	    }
	}
    }
    else if ((ep = hsearch(server_entry, FIND)) != NULL) {
	SERVER *server;
	server = (SERVER *)(ep->data);
	if (open_stpserver_at (server->hostname, server->port) == 0) return (0);
	err(FATAL, "STP: cannot connect to server %s\n", name);
    }
    else {
	fprintf (stderr, "Error: Unknown group or server: %s\n", name);
	return (-1);
    }
    err(FATAL, "STP: cannot connect to any server in group %s\n", name);
    return (-1);
}

/************************************************************************/
/*  open_stpserver_at:							*/
/*	Open a connection to STP server at specified host and port.	*/
/*	Return 0 on success, non-zero on failure.			*/
/*	Exit on fatal error.						*/
/************************************************************************/
int open_stpserver_at (char *hostname, int port)
{
    int two;
    FILE *fdopen();
    struct sockaddr_in servaddr, cliaddr;
    struct hostent *hp;
    char line[TOKENLEN];
    /*socklen_t len;*/
    uint len;
    unsigned long  ipadd;
    int status;

    if (isalpha(hostname[0])) {
	hp = gethostbyname(hostname);	/* probe name server by name */
    } 
    else {
	ipadd = inet_addr(hostname);	/* convert IP to binary format */
	hp = gethostbyaddr((char*)&ipadd, sizeof(ipadd), AF_INET);	
					/*probe name server by address */
    }
    if (hp == NULL) {
	err(WARN,"STP:  Unable to resolve address for %s\n", hostname);
	return (-1);
    }
	
    /* set up a socket */
    if ( (ws_sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	err(WARN,"Wave_Server: socket connect error\n");
    bzero(&servaddr, sizeof(servaddr));
    servaddr.sin_family = hp->h_addrtype;
    servaddr.sin_port   = htons((unsigned short int)port);
    memcpy(&servaddr.sin_addr,hp->h_addr, hp->h_length);
	  
    /* connect socket to Wave Server Port */
    status = connect(ws_sockfd, (struct sockaddr *) &servaddr, sizeof(servaddr));
    if (status < 0) return (-1);
    len = sizeof(cliaddr);
    if (getsockname(ws_sockfd, (struct sockaddr *) &cliaddr, (int *)(&len)) < 0)
	err(WARN,"STP: getsockname error");
	  
    ws_fdr= fdopen(ws_sockfd,"r");
    ws_fdw= fdopen(ws_sockfd,"w");
    if( ws_fdr == NULL || ws_fdw == NULL)
	err(FATAL,"STP: cannot open ws_fdw/ws_fdr\n");

    /* Copy server info to global structure for later use. */
    iserv = 0;
    strcpy (address[iserv].name, hostname);
    strcpy (address[iserv].addr, inet_ntoa(servaddr.sin_addr));
    address[iserv].port = port;

	  /* send authorization */
	  sprintf(line,"%s %s %s %s\n","STP",PASSWD,VERSION, "stpc");
	  ws_putline(line);
	  fflush(ws_fdw);
	  
	  /* get connection acknowledgement */
	  ws_getline(line,MAXLINE);
	  if(strcmp(line,"CONNECTED\n") != 0) {
	    err(WARN,"STP: Connection rejected by waveserver - %s\n", line);
	    return (-1);
	  }

	  fprintf(stderr,"STP: Connected to %s (%s)\n", address[iserv].name, address[iserv].addr);
	  
	  /* send a binary sample so the server can establish byte order */
	  two= 2;
	  ws_putdata(&two,4);
	  fflush(ws_fdw);
	  
	  /* receiver possible message from server */
	  receive_data();

    return (0);
}
